/*
    reference alt-inflator
*/

#pragma once
#include <stdint.h>
#include <cstring>
#include <intrin.h>
#include <iostream>
#include <windows.h>

double PCFreq = 0.0;
__int64 CounterStart = 0;

void StartCounter()
{
    LARGE_INTEGER li;
    QueryPerformanceFrequency(&li);

    PCFreq = double(li.QuadPart)/1000000.0;

    QueryPerformanceCounter(&li);
    CounterStart = li.QuadPart;
}
double GetCounter()
{
    LARGE_INTEGER li;
    QueryPerformanceCounter(&li);
    return double(li.QuadPart-CounterStart)/PCFreq;
}

/* format:
    * blocks are dword-aligned
    * first byte: 
    *   if 0, end of stream
    *   otherwise:
    * first dword (little endian):
    *   bit0: 1
    *   bit1: is stored
    *   rest: length << 2
    * if not stored:
    *   dword: decompressed length of block
    *   { codebook }
    *   symbols
    * codebook:
    *   bit[16] ispresent: 0 = no symbols in range are used, 1 = some may be used
    *   bit[16] isdefault: 0 = subrange has a mask, 1 = subrange is completely used
    *   { bit[16]: subrange }
    *   bit[16]: length codes 0x100 - 0x10F
    *   bit[16]: length codes 0x110 - 0x11F
    *   { byte:5 symbollength; byte:3 repeatcode [byte repeatopt] }
    * symbols:
    *   .align 4
    *   dwords (little endian) contain symbols packed 
    *   codes 0 - 255 correspond to those bytes
    *   codes 0x100 - 0x11F correspond to matches
    *   code  0x120 is "End Of Stream"
    *   
    */

extern "C" void decode(uint8_t* outptr, uint8_t* inptr);
extern "C" void decode_nobmi(uint8_t* outptr, uint8_t* inptr);
extern "C" uint8_t mBsrShiftTable[32];
extern "C" uint32_t mBsrCodeTable[1024];
extern "C" uintptr_t mBsrLenTable[32];

uint32_t BSR(uint32_t x)
{
    return __lzcnt(x) ^ 31;
}

uint32_t alt_inflate(uint8_t* outptr, uint8_t* inptr)
{
    int outputlength = 0;

    uint32_t total_decompressed_size = 0;

    while (*inptr)
    {
        uint32_t header = *reinterpret_cast<uint32_t*>(inptr);
        inptr += 4;
        size_t blocklen = header >> 2;
        if (header & 2)
        {
            std::memcpy(outptr, inptr, blocklen);
            outptr += blocklen;
            inptr += blocklen;
            outputlength += blocklen;
        }
        else
        {
            uint32_t decompressed_size = *reinterpret_cast<uint32_t*>(inptr);
            total_decompressed_size += decompressed_size;
            inptr += 4;
            // read codebook(s)

            // parse sparse bit array
            uint32_t cbheader = *reinterpret_cast<uint32_t*>(inptr);
            inptr += 4;
            uint16_t symbolspresent[19];
            for (int subrange = 0; subrange < 16; subrange++)
            {
                if (cbheader & (1 << subrange))
                {
                    if (cbheader & (0x10000 << subrange))
                        symbolspresent[subrange] = ~0;
                    else
                    {
                        symbolspresent[subrange] = *reinterpret_cast<uint16_t*>(inptr);
                        inptr += 2;
                    }
                }
                else symbolspresent[subrange] = 0;
            }
            symbolspresent[16] = *reinterpret_cast<uint16_t*>(inptr);
            inptr += 2;
            symbolspresent[17] = *reinterpret_cast<uint16_t*>(inptr);
            inptr += 2;
            symbolspresent[18] = 1;

            // parse symbol lengths
            const int totalsymbols = 256 + 32 + 1;
            char symbollength[totalsymbols];
            std::memset(symbollength, 0, sizeof(symbollength));

            for (int i = 0; i < totalsymbols; i++)
            {
                if (symbolspresent[i >> 4] & (1 << (i & 15)))
                {
                    int temp = *inptr++;
                    int repcount = temp >> 5;
                    if (repcount == 0)
                        repcount = 8 + *inptr++;
                    int codelength = temp & 31;
                    for (int j = 0; j < repcount && i < totalsymbols; j++, i++)
                    {
                        if (symbolspresent[i >> 4] & (1 << (i & 15)))
                            symbollength[i] = codelength;
                        else
                            j--;
                    }
                    i--;
                }
            }

            // create decoding tables
            for (int i = 0; i < 32; i++)
                mBsrShiftTable[i] = 32;

            uint32_t total = 0;
		    uint32_t base = 0;
		    uint8_t *decdst = reinterpret_cast<uint8_t*>(mBsrCodeTable);
		    uint8_t *declimit = reinterpret_cast<uint8_t*>(mBsrCodeTable) + sizeof mBsrCodeTable;

		    int curdecidx = -1;
            for (uint32_t len = 32; len >= 1; --len)
            {
                uint32_t inc = 0x80000000 >> (len - 1);
    			uint32_t prevBase = base;

                for(int j = 0; j < totalsymbols; j++)
                {
                    if (symbollength[j] == len)
                    {
					    base += inc;
                        int decidx = BSR((base - 1) | 1);
					    int basedecidx = curdecidx;
					    while(curdecidx < decidx)
                        {
						    if (curdecidx >= 0)
                            {
							    int decrepeat = 1 << (32 - mBsrShiftTable[curdecidx] - len);
							    while(decrepeat--) {
								    *decdst++ = len;
                                    *decdst++ = j;
                                    *decdst++ = j >> 8;
                                    *decdst++ = 0;
                                }
						    }

						    ++curdecidx;
						    uint32_t shift = mBsrShiftTable[curdecidx];
						    if (shift > 32 - len)
                            {
							    shift = 32 - len;
							    mBsrShiftTable[curdecidx] = 32 - len;
							    mBsrLenTable[curdecidx] = (uintptr_t)decdst - 4*((1U << curdecidx) >> shift);
						    }
                        }

					    int decrepeat = 1 << (32 - mBsrShiftTable[curdecidx] - len);
						while(decrepeat--) {
							*decdst++ = len;
                            *decdst++ = j;
                            *decdst++ = j >> 8;
                            *decdst++ = 0;
                        }
                    }
                }
            }

            // just for benchmarking, decompress 200 times
            for (int i = 0; i < 200; i++)
            {
                StartCounter();

                decode_nobmi(outptr, inptr);

                double time = GetCounter();
                std::cout << "time: " << time << std::endl;
            }
        }

        return total_decompressed_size;
    }

    return outputlength;
}