﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;

namespace Compress
{
    class Program
    {
        class Node
        {
            public uint XYZ;
            public int pos;
            public Node next;

            public Node(uint XYZ, int pos, Node next)
            {
                this.XYZ = XYZ;
                this.pos = pos;
                this.next = next;
            }
        }

        class HNode
        {
            public uint W;
            public HNode L, R;
            public uint S;

            public HNode(uint w, HNode l, HNode r, uint s)
            {
                W = w;
                L = l;
                R = r;
                S = s;
            }
        }

        const uint dictsize = 65521;
        const int chainlimit = 1 << 11;

        static void Main(string[] args)
        {
            byte[] data = File.ReadAllBytes(@"C:\Users\Harold\Downloads\alice29.txt");
            List<int> symbols = new List<int>(1024);
            symbols.Add(data[0]);
            Node[] m = new Node[dictsize];
            int[] hashchainlength = new int[dictsize];
            int[] lengthoffsets = { 3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31, 35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 259, 323, 387, 451, int.MaxValue };
            int[] extralengthbits = { 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6 };
            int[] distoffsets = { 1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193, 257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145, 8193, 12289, 16385, 24577, 32769, 49153, int.MaxValue };
            int[] extradistbits = { 0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 14, 14 };

            {
                uint xyz = (uint)(data[0] | (data[1] << 8) | (data[2] << 16));
                m[xyz % dictsize] = new Node(xyz, 0, null);
                hashchainlength[xyz % dictsize] = 1;
            }
            for (int i = 1; i < data.Length; i++)
            {
                // find earlier
                if ((i + 2) < data.Length)
                {
                    uint xyz = (uint)(data[i] | (data[i + 1] << 8) | (data[i + 2] << 16));
                    Node c = m[xyz % dictsize];
                    int bestmatchpos = -1;
                    int bestmatchlength = 0;
                    int index = 0;
                    while (c != null)
                    {
                        index++;
                        if (i - c.pos > 0x10000)
                        {
                            c.next = null;
                            hashchainlength[xyz % dictsize] = index;
                            break;
                        }
                        if (xyz == c.XYZ)
                        {
                            int j = c.pos;
                            int length = 3;
                            while (i + length < data.Length && data[j + length % (i - j)] == data[i + length] && length < 515)
                                length++;
                            if (length > bestmatchlength)
                            {
                                bestmatchlength = length;
                                bestmatchpos = j;
                            }
                        }
                        c = c.next;
                    }
                    if (hashchainlength[xyz % dictsize]++ > chainlimit)
                    {
                        hashchainlength[xyz % dictsize] = 0;
                        m[xyz % dictsize] = null;
                    }
                    m[xyz % dictsize] = new Node(xyz, i, m[xyz % dictsize]);
                    if (bestmatchlength == 0)
                        symbols.Add(data[i]);
                    else
                    {
                        int lengthindex = Array.BinarySearch<int>(lengthoffsets, bestmatchlength);
                        if (lengthindex < 0)
                            lengthindex = ~lengthindex - 1;
                        symbols.Add(lengthindex + 256);
                        symbols.Add(bestmatchlength);
                        symbols.Add(i - bestmatchpos);

                        if (i - bestmatchpos > 0x10000)
                            throw new Exception();

                        for (int j = i; j < i + bestmatchlength && (j + 2) < data.Length; j++)
                        {
                            uint _xyz = (uint)(data[j] | (data[j + 1] << 8) | (data[j + 2] << 16));
                            if (hashchainlength[_xyz % dictsize]++ > chainlimit)
                            {
                                hashchainlength[_xyz % dictsize] = 0;
                                m[_xyz % dictsize] = null;
                            }
                            m[_xyz % dictsize] = new Node(_xyz, j, m[_xyz % dictsize]);
                        }
                        i += bestmatchlength - 1;
                    }
                }
                else symbols.Add(data[i]);
            }

            List<byte> dec = new List<byte>(1024);
            for (int i = 0; i < symbols.Count; i++)
            {
                int symbol = symbols[i];
                if (symbol < 256)
                    dec.Add((byte)symbol);
                else
                {
                    int length = symbols[i + 1];
                    int dist = symbols[i + 2];
                    i += 2;
                    int pos = dec.Count - dist;
                    for (int j = 0; j < length; j++)
                    {
                        dec.Add(dec[pos + j]);
                    }
                }
            }

            int errors = 0;
            for (int i = 0; i < data.Length; i++)
            {
                if (data[i] != dec[i])
                {
                    Console.WriteLine("{0} - {1}", data[i], dec[i]);
                    errors++;
                }
            }

            if (errors == 0)
                Console.WriteLine("Decoded symbols without errors");
            else
                throw new Exception();

            dec = null;


            //symbols.Clear();
            //symbols.AddRange(new int[] { 4, 5, 5, 4, 0, 1, 2, 3, 4, 5, 5 });
            symbols.Add(0x120);


            uint[] histogram = new uint[0x121];
            histogram[0x120] = 1;
            for (int i = 0; i < symbols.Count; i++)
            {
                int symbol = symbols[i];
                histogram[symbol]++;
                if (symbol >= 256)
                    i += 2;
            }

            int[] symbollengths = new int[histogram.Length];

            do
            {
                uint[] symbols_by_weight = new uint[histogram.Length];
                for (uint i = 0; i < symbols_by_weight.Length; i++)
                    symbols_by_weight[i] = i;
                Array.Sort<uint>(symbols_by_weight, (uint i0, uint i1) => histogram[i0].CompareTo(histogram[i1]));

                Queue<HNode> leafq = new Queue<HNode>(0x121);
                foreach (var s in symbols_by_weight)
                {
                    if (histogram[s] != 0)
                        leafq.Enqueue(new HNode(histogram[s], null, null, s));
                }

                Queue<HNode> internalq = new Queue<HNode>(0x121);
                {
                    HNode l = leafq.Dequeue();
                    HNode r = leafq.Dequeue();
                    internalq.Enqueue(new HNode(l.W + r.W, l, r, 0));
                }

                while (leafq.Count + internalq.Count > 1)
                {
                    HNode l = leafq.Count == 0 || internalq.Peek().W < leafq.Peek().W ? internalq.Dequeue() : leafq.Dequeue();
                    HNode r = leafq.Count == 0 || (internalq.Count != 0 && internalq.Peek().W < leafq.Peek().W) ? internalq.Dequeue() : leafq.Dequeue();
                    internalq.Enqueue(new HNode(l.W + r.W, l, r, 0));
                }

                FillLengths(symbollengths, internalq.Dequeue(), 0);

                if (symbollengths.Max() > 31)
                {
                    for (int i = 0; i < histogram.Length; i++)
                    {
                        if (histogram[i] != 0)
                            histogram[i] = Math.Max(1, histogram[i] >> 1);
                    }
                    continue;
                }
                

            } while (false);

            //create encoding tables
            const int maxcodelength = 31;
            uint[] bl_count = new uint[maxcodelength + 1];;
            for (int i = 0; i < symbollengths.Length; i++)
                bl_count[symbollengths[i]]++;
            uint code = 0;
            uint[] next_code = new uint[maxcodelength + 1];
            next_code[maxcodelength] = 0;
            for (int bits = maxcodelength - 1; bits >= 0; bits--)
            {
                code = (code + bl_count[bits + 1]) >> 1;
                next_code[bits] = code;
            }

            uint[] codes = new uint[histogram.Length];
            for (int n = 0; n < codes.Length; n++)
            {
                int len = symbollengths[n];
                if (len != 0)
                {
                    codes[n] = next_code[len];
                    next_code[len]++;
                }
            }

            File.Delete(@"C:\Users\Harold\Documents\out.az");
            using (FileStream fs = File.OpenWrite(@"C:\Users\Harold\Documents\out.az"))
            {
                BinaryWriter w = new BinaryWriter(fs);

                w.Write(1);
                w.Write(data.Length);
                ushort[] subranges = new ushort[18];                
                for (int i = 0; i < 0x120; i++)
                {
                    if (symbollengths[i] != 0)
                        subranges[i >> 4] |= (ushort)(1 << (i & 15));
                }
                uint cbheader = 0;
                for (int i = 0; i < 16; i++)
                {
                    if (subranges[i] != 0)
                    {
                        if (subranges[i] == ushort.MaxValue)
                            cbheader |= 0x10001U << i;
                        else
                            cbheader |= 1U << i;
                    }
                }
                w.Write(cbheader);
                int written = 0;
                for (int i = 0; i < 16; i++)
                {
                    if (subranges[i] != 0 && subranges[i] != ushort.MaxValue)
                    {
                        w.Write(subranges[i]);
                        written += 2;
                    }
                }
                w.Write(subranges[16]);
                w.Write(subranges[17]);

                for (int i = 0; i < symbollengths.Length; )
                {
                    if (symbollengths[i] != 0)
                    {
                        int slen = symbollengths[i];
                        int repcount = 1;
                        int j = i + 1;
                        while (j < symbollengths.Length && symbollengths[j] == 0) j++;
                        while (j < symbollengths.Length && slen == symbollengths[j] && repcount < 263)
                        {
                            j++; repcount++;
                            while (j < symbollengths.Length && symbollengths[j] == 0) j++;
                        }
                        if (repcount < 8)
                        {
                            w.Write((byte)((repcount << 5) | slen));
                            written++;
                        }
                        else
                        {
                            w.Write((byte)slen);
                            w.Write((byte)(repcount - 8));
                            written += 2;
                        }
                        i = j;
                    }
                    else i++;
                }

                int align = ((written + 3) & -4) - written;
                w.Write(new byte[align]);

                ulong buffer = 0;
                int len = 0;

                for (int i = 0; i < symbols.Count; i++)
                {
                    int symbol = symbols[i];
                    int l = symbollengths[symbol];
                    len += l;
                    buffer = (buffer << l) | codes[symbol];
                    if (len >= 32)
                    {
                        w.Write((uint)(buffer >> (len - 32)));
                        len -= 32;
                    }
                    if (symbol >= 256)
                    {
                        if (symbol == 0x120)
                            break;
                        int length = symbols[i + 1];
                        int dist = symbols[i + 2];
                        i += 2;

                        int length_extra_bits = extralengthbits[symbol & 0xFF];
                        if (length_extra_bits != 0)
                        {
                            int length_diff = length - lengthoffsets[symbol & 0xFF];
                            len += length_extra_bits;
                            buffer = (buffer << length_extra_bits) | (uint)length_diff;
                            if (len >= 32)
                            {
                                w.Write((uint)(buffer >> (len - 32)));
                                len -= 32;
                            }
                        }

                        int distindex = Array.BinarySearch<int>(distoffsets, dist);
                        if (distindex < 0)
                            distindex = ~distindex - 1;
                        int dist_extra_bits = extradistbits[distindex];
                        int dist_diff = dist - distoffsets[distindex];
                        uint distsymbol = (uint)dist_diff | ((uint)distindex << dist_extra_bits);
                        int distsymlen = dist_extra_bits + 5;
                        len += distsymlen;
                        buffer = (buffer << distsymlen) | distsymbol;
                        if (len >= 32)
                        {
                            w.Write((uint)(buffer >> (len - 32)));
                            len -= 32;
                        }
                    }
                }

                if (len != 0)
                    w.Write((uint)(buffer << (32 - len)));

                w.Flush();
                w.Close();
            }

            return;
        }

        static void FillLengths(int[] lengths, uint[] internalnodes, int depth, uint node)
        {
            depth++;
            uint left = node & 0xFFFF;
            uint right = node >> 16;

            if (left <= 0x7FFF)
                FillLengths(lengths, internalnodes, depth, internalnodes[left]);
            else
                lengths[left & 0x7FFF] = depth;

            if (right <= 0x7FFF)
                FillLengths(lengths, internalnodes, depth, internalnodes[right]);
            else
                lengths[right & 0x7FFF] = depth;
        }

        static void FillLengths(int[] lengths, HNode root, int depth)
        {
            HNode l = root.L;
            HNode r = root.R;
            if (l != null)
            {
                FillLengths(lengths, l, depth + 1);
                FillLengths(lengths, r, depth + 1);
            }
            else
                lengths[root.S] = depth;
        }
    }
}
